unit ListOOUN;

  (*--------------------------------------------------------------------------*)
  (*                                                                          *)
  (* TListOO: Version of TListO adding Ownership, Uniqueness and NoNil. Also  *)
  (*          adds Sorted property to keep list in sorted state.              *)
  (* Version:       990904.1623                                               *)
  (*                                                                          *)
  (* Copyright:  Jim O'Flaherty, Jr. - 1999                                   *)
  (* Created On: 1999 Sep 02                                                  *)
  (*                                                                          *)
  (*--------------------------------------------------------------------------*)

  interface

    uses
      Classes, ListO;

    type
      TListOOUN = class(TListO)
        protected
          FOwns,
          FUnique,
          FNoNils,
          FSorted: Boolean;
          FSortCompare: TListOSortCompare;
          procedure DoSort; virtual;
          procedure SetCapacity(iWhat: Integer); override;
          procedure SetCount(iWhat: Integer); override;
          procedure SetItems(iWhat: Integer; oWhat: TObject); override;
          function GetOwns: Boolean; virtual;
          function GetUnique: Boolean; virtual;
          procedure SetUnique(bWhat: Boolean); virtual;
          function GetNoNils: Boolean; virtual;
          procedure SetNoNils(bWhat: Boolean); virtual;
          function GetSorted: Boolean; virtual;
          procedure SetSorted(bWhat: Boolean); virtual;
        public
          constructor CreateO; virtual;
             //Owns set True.
          constructor CreateOUN(bOwns, bUnique, bNoNils: Boolean); virtual;
              //Each set according to parameter.
          constructor CreateOUNSet; virtual;
              //All set True.
          destructor Destroy; override;

          function Add(oWhat: TObject): Integer; override;
          procedure Clear; override;
          procedure Delete(iWhat: Integer); override;
          procedure Exchange(iWhat1, iWhat2: Integer); override;
          function Find(oWhat: TObject): Integer; virtual;
          function FindInRange(oWhat: TObject; iStartingAt, iEndingAt: Integer): Integer; virtual;
          procedure Insert(iWhat: Integer; oWhat: TObject); override;
          procedure Move(iWhat1, iWhat2: Integer); override;
          function Remove(oWhat: TObject): Integer; override;

          property Owns: Boolean read GetOwns;
              //If True, list managing freeing of held objects.
              //    Only allowed to be set at instance creation.
          property Unique: Boolean read GetUnique write SetUnique;
              //If True, list rejects attempts to re-enter same object.
              //    If Sorted is True, the search for uniqueness is optimal
              //    else the search is linear.
          property NoNils: Boolean read GetNoNils write SetNoNils;
              //If True, all operation resulting in leaving the value nil
              //    accessible in an operation is rejected.
          property Sorted: Boolean read GetSorted write SetSorted;
              //If True, ensures list is sorted after *any* modification.
          property Compare: TListOSortCompare read FSortCompare write FSortCompare;
              //Method used by the Find and Sort methods.
        end;

  implementation
//protected
    procedure TListOOUN.DoSort;
      begin
      if Sorted then
        Sort(Compare);
      end;

    procedure TListOOUN.SetCapacity(iWhat: Integer);
      var
        i: Integer;
      begin
      if iWhat < Count then
        if Owns then
          for i := iWhat to LastIndex do
            Items[i].Free; //In place to optimize call to inherited SetCount.
      inherited SetCapacity(iWhat);
      end;

    procedure TListOOUN.SetCount(iWhat: Integer);
      var
        i: Integer;
      begin
      if iWhat < Count then
        begin
        if Owns then
          for i := iWhat to LastIndex do
            Items[i].Free; //In place to optimize call to inherited SetCount.
        end
      else if (iWhat > Count) then
        begin
        if NoNils then
          Error('TListOOUN.SetCount: NoNils is On', 0)
        else if Unique then
          Error('TListOOUN.SetCount: Unique is On', 0);
        end;
      inherited SetCount(iWhat);
      end;

    procedure TListOOUN.SetItems(iWhat: Integer; oWhat: TObject);
      begin
      if NoNils and (oWhat = nil) then
        Error('TListOOUN.SetItems: NoNils is On', iWhat);
      if Unique and (IndexOf(oWhat) > -1) then
        Error('TListOOUN.SetItems: Unique is On', iWhat);
      if Owns and (iWhat < Count) then
        if Items[iWhat] <> nil then
          Items[iWhat].Free;
      inherited SetItems(iWhat, oWhat);
      DoSort;
      end;

    function TListOOUN.GetOwns: Boolean;
      begin
      Result := FOwns;
      end;

    function TListOOUN.GetUnique: Boolean;
      begin
      Result := FUnique;
      end;

    procedure TListOOUN.SetUnique(bWhat: Boolean);
      var
        i,
        j: Integer;
      begin
      if FUnique <> bWhat then
        begin
        if bWhat then
          for i := 0 to LastIndex do
            for j := i + 1 to LastIndex do
              if Items[i] = Items[j] then
                Error('TListOOUN.SetUnique: Found a duplicate reference in Items[]', 0);
        FUnique := bWhat;
        end;
      end;

    function TListOOUN.GetNoNils: Boolean;
      begin
      Result := FNoNils;
      end;

    procedure TListOOUN.SetNoNils(bWhat: Boolean);
      var
        i: Integer;
      begin
      if FNoNils <> bWhat then
        begin
        if bWhat then
          for i := 0 to LastIndex do
            if Items[i] = nil then
              Error('TListOOUN.SetNoNils: Found a entry in Items[]', 0);
        FNoNils := bWhat;
        end;
      end;

    function TListOOUN.GetSorted: Boolean;
      begin
      Result := FSorted;
      end;

    procedure TListOOUN.SetSorted(bWhat: Boolean);
      begin
      if FSorted <> bWhat then
        if (not bWhat) or Assigned(Compare) then
          begin
          FSorted := bWhat;
          DoSort;
          end
        else
          Error('TListOOUN.SetSorted: Compare is not Assigned', 0);
      end;
//public
    constructor TListOOUN.CreateO;
      begin
      Create;
      FOwns := True;
      end;

    constructor TListOOUN.CreateOUN(bOwns, bUnique, bNoNils: Boolean);
      begin
      Create;
      FOwns := bOwns;
      FUnique := bUnique;
      FNoNils := bNoNils;
      end;

    constructor TListOOUN.CreateOUNSet;
      begin
      Create;
      FOwns := True;
      FUnique := True;
      FNoNils := True;
      end;

    destructor TListOOUN.Destroy;
      begin
      Sorted := False;
      Clear;
      inherited Destroy;
      end;

    function TListOOUN.Add(oWhat: TObject): Integer;
      begin
      if Unique then
        if IndexOf(oWhat) > -1 then
          Error('TListOOUN.Add: Unique is on', 0);
      if NoNils then
        if oWhat = nil then
          Error('TListOOUN.Add: NoNils is no', 0);
      Result := inherited Add(oWhat);
      DoSort;
      if Sorted then
        Result := IndexOf(oWhat);
      end;

    procedure TListOOUN.Clear;
      var
        i: Integer;
      begin
      if Owns then
        for i := 0 to LastIndex do
          Items[i].Free;
      inherited Clear;
      end;

    procedure TListOOUN.Delete(iWhat: Integer);
      begin
      if Owns then
        Items[iWhat].Free;
      inherited Delete(iWhat);
      end;

    procedure TListOOUN.Exchange(iWhat1, iWhat2: Integer);
      begin
      if Sorted then
        Error('TListOOUN.Exchange: Sorted is On', iWhat1);
      inherited Exchange(iWhat1, iWhat2);
      end;

    function TListOOUN.Find(oWhat: TObject): Integer;
      begin
      Result := FindInRange(oWhat, 0, LastIndex);
      end;

    function TListOOUN.FindInRange(oWhat: TObject; iStartingAt, iEndingAt: Integer): Integer;
      var
        iStart,
        iEnd,
        iCR: Integer;
      begin
      if (not ((iStartingAt >= 0) and (iStartingAt <= iEndingAt))) then
        Error('TListOOUN.FindInRange: iStartingAt not in valid range of 0..iEndingAt', iStartingAt);
      if (not ((iEndingAt >= iStartingAt) and (iEndingAt <= LastIndex))) then
        Error('TListOOUN.FindInRange: iEndingAt not in valid range of iStartingAt..LastIndex', iEndingAt);
      if Sorted then
        begin
        //Use Binary search to locate item.
        iStart := iStartingAt;
        iEnd := iEndingAt;
        repeat
          Result := iStart + ((iEnd - iStart) div 2);
          iCR := Compare(oWhat, Items[Result]);
          if iCR > 0 then
            iStart := Result + 1
          else if iCR < 0 then
            iEnd := Result - 1;
        until ((iCR = 0) or (iStart > iEnd));
        if iCR <> 0 then
          Result := -Result; //abs(Result) is sorted Insertion position.
        end
      else
        begin
        //Use a linear search to locate item.
        Result := iStartingAt;
        while (Result <= iEndingAt) and (Compare(oWhat, Items[Result]) <> 0) do
          Inc(Result);
        if Result > iEndingAt then
          Result := -Count;
        end;
      end;

    procedure TListOOUN.Insert(iWhat: Integer; oWhat: TObject);
      begin
      if Sorted then
        Error('TListOOUN.Insert: Sorted is On', iWhat);
      inherited Insert(iWhat, oWhat);
      end;

    procedure TListOOUN.Move(iWhat1, iWhat2: Integer);
      begin
      if Sorted then
        Error('TListOOUN.Move: Sorted is On', iWhat1);
      inherited Move(iWhat1, iWhat2);
      end;

    function TListOOUN.Remove(oWhat: TObject): Integer;
      begin
      Result := inherited Remove(oWhat);
      if Owns then
        oWhat.Free;
      end;

  end.
